import re
import sys


def evaluate_success(control_results, case_results, factor = 0.001):
    success = True
    try:
        for key in case_results:
            if case_results[key] >= control_results[key] * factor:
                success = False
    except TypeError:
        success = False
    return success


def hash_from_list(_list):
    _list = list(sorted(_list))
    val = (sys.maxsize + 1) * 2
    return hash(tuple(_list)) + val


def generate_jobscript(template, parameters):
    with open(template) as f:
        job = f.read()
        for parameter, value in parameters.items():
            job = job.replace("%%%s%%" % parameter, value)
    return job


def submit_job(job, remote_jobscript, communicator):
        with communicator.sftp.file(remote_jobscript, "w") as jobscript:
            jobscript.write(job)

        command = "sbatch %s" % remote_jobscript
        stdin, stdout, stderr = communicator.execute(command)
        output = str(stdout.read())

        jobid_pattern = "Submitted batch job +([0-9]+)"
        jobid = re.search(jobid_pattern, output).group(1)
        return jobid


def job_status(job_id, communicator):
    command = 'sacct --job %s --format="State,Elapsed"' % job_id
    stdin, stdout, stderr = communicator.execute(command)
    output = stdout.read().decode()
    try:
        split_output = output.split("\n")[2].split()
        status = split_output[0]
        time = split_output[1]
    except IndexError:
        status = "PENDING"
        time = "00:00"
    return status, time


def obtain_result(remote_rundir, communicator):
    result = {}
    with communicator.sftp.open("%s/accuracy.txt" % remote_rundir) as f:
        data = [l for l in f]
        for l in data:
            elements = l.split()
            if len(elements) == 1:
                result["error"] = float(elements[0])
            elif len(elements) == 2:
                key = elements[0].strip()
                value = float(elements[1])
                result[key] = value
            else:
                raise AssertionError("accuracy.txt file has not the proper format")

        return result


def generate_namelist(variable_set, filename, _vault, reduced_precision_level=10):
    variables = [v for v in _vault.variables if v.id is not None]
    for index, v in enumerate(variables):
        if v.id in variable_set:
            v.precision = reduced_precision_level
        else:
            v.precision = 52
    _vault.generate_namelist_precisions(filename=filename, exclude_double_precision=True)


def var_set_identifier(var_set):
    to_return = ""
    current_var_set = var_set
    while current_var_set.parent is not None:
        if len(to_return) > 100:
            raise AssertionError("Something strange happens here!")
        to_return += str(current_var_set.child_number)
        current_var_set = current_var_set.parent
    to_return = "l%02s: %s" % (str(var_set.level), to_return[::-1])
    return to_return


def divide_set(set0):
    assert len(set0) > 1
    set1 = []
    set2 = set0[:]
    for i in range(int(len(set0)/2)):
        set1.append(set2.pop(0))
    return set1, set2


class Job(object):
    def __init__(self, variable_set, reduced_precision_level=10, communicator=None, vault=None, template="Templates/singleJobscript.sh"):
        # Init variables
        self.vault = vault
        self.variable_set = variable_set
        self.communicator = communicator
        self.job_name = "BinaryRefinement%i" % self.hash
        self.remote_scratch = self.communicator.remote_scratch
        self.remote_logdir = "%s/LOGS" % self.remote_scratch
        self.remote_namelist_dir = "%s/Namelists" % self.remote_logdir

        self.remote_rundir = "%s/RefinementTest%i" % (self.remote_scratch, self.hash)

        self.remote_jobscript_path = "%s/refinement_%i.cmd" % (self.remote_logdir, self.hash)
        self.remote_namelist = "%s/refinement_namelist_%i" % (self.remote_namelist_dir, self.hash)
        self.local_namelist = "NewWorlds/namelist_precisions"

        self.template = template
        # parameters to replace:
        self.parameters = {"JOBNAME": self.job_name,
                           "RUNDIR": self.remote_rundir,
                           "LOGDIR": self.remote_logdir,
                           "NAMELIST_PATH": self.remote_namelist,
                           "NUMPROC": "48"
                           }

        self.jobscript = generate_jobscript(self.template, self.parameters)

        self.elapsed_time = None
        self.job_id = None
        self._result = None

        self.reduced_precision_level = reduced_precision_level

        self.assertion_job = None

        # The default status is PENDING
        self._status = "PENDING"

    @property
    def remote_status(self):
        _status, _time = job_status(self.job_id, self.communicator)
        _status = _status.strip()
        return _status

    @property
    def status(self):
        return self._status
    
    @status.setter
    def status(self, new_value):
        possible_status = ["PENDING", "RUNNING", "ASSERTION_PENDING", "SUCCESS", "FAILED", "SUSPENDED"]
        # The list of possible status shall cover all the possible situations in which a job can be found.
        # PENDING: The job has been created and ready to be launched.
        # A PENDING JOB CAN ONLY HAVE SUSPENDED ANCESTORS AND SUCCESS OR FAILED DESCENDANTS
        # RUNNING: The job is RUNNING.
        # A RUNNING JOB CAN ONLY HAVE SUSPENDED ANCESTORS AND SUCCESS OR FAILED DESCENDANTS
        # ASSERTION_PENDING: A Job that has been completed with successful results but still needs assertion.
        # A ASSERTION_PENDING JOB CAN ONLY HAVE SUSPENDED ANCESTORS AND SUCCESS OR FAILED DESCENDANTS
        # SUCCESS: A job that has been completed with successful results and asserted.
        # A SUCCESS job can not have both descendants FAILED
        # FAILED: A job that has been completed with unsuccessful results or a failed assertion and that can not be
        # split ( Already was or a single variable set).
        # SUSPENDED: A job that has been completed with unsuccessful results that has been split and depends on the
        # results of descendant jobs.
        # SUSPENDED ancestors must be SUSPENDED.
        # At least one direct descendant should be different from FAILED or COMPLETED
        if new_value in possible_status:
            self._status = new_value
        else:
            raise AssertionError("Unknown status")

    # @property
    def result(self):
        try:
            self._result = obtain_result(self.remote_rundir, self.communicator)
            return self._result
        except FileNotFoundError:
            # time.sleep(1)
            # return self.result(level=level+1)
            return None

    @property
    def hash(self):
        return hash_from_list(self.variable_set)

    def recover_job_id(self):
        job_id_filename = "%s/job_id.txt" % self.remote_rundir
        self.job_id = self.communicator.sftp.open(job_id_filename).read().decode().strip()

    def check_if_result_already_exists(self):
        self.update_parameters()
        accuracy_filename = "%s/accuracy.txt" % self.remote_rundir
        try:
            self.communicator.sftp.open(accuracy_filename)
            # Recovering job_id
            self.recover_job_id()
            return True
        except IOError:
            return False

    def check_if_already_submitted(self):
        self.update_parameters()
        try:
            command = 'squeue --format="%i %j"'
            stdin, stdout, stderr = self.communicator.execute(command)

            output = stdout.read().decode()
            lines = output.split("\n")
            if len(lines) > 1:
                jobs_in_remote_queue = lines[1:]
                for line in jobs_in_remote_queue:
                    if line.strip():
                        job_id, job_name = line.split()
                        if job_name == self.job_name:
                            return job_id

        except IOError:
            return None
        return None

    def submit(self, force=False):
        # Check if result already exists
        if force is False:
            completed = self.check_if_result_already_exists()
        else:
            completed = False
        if not completed:
            self.update_parameters()
            # Generate namelist and put it in the remote system
            generate_namelist(self.variable_set, self.local_namelist, self.vault, self.reduced_precision_level)
            self.communicator.put(self.local_namelist, self.remote_namelist)
            # Submit job and obtain job id
            if force is False:
                self.job_id = self.check_if_already_submitted()
            else:
                self.job_id = None
            if self.job_id is None:
                self.job_id = submit_job(self.jobscript, self.remote_jobscript_path, self.communicator)
                print("Job %i successfully submitted to the remote machine with job_id %s" % (self.hash, self.job_id))
            else:
                print("Job %i was already in queue with job_id %s" % (self.hash, self.job_id))
        else:
            pass

    def update_parameters(self):
        self.remote_rundir = "%s/RefinementTest%i" % (self.remote_scratch, self.hash)
        self.job_name = "BinaryRefinement%i" % self.hash

        self.remote_jobscript_path = "%s/refinement_%i.cmd" % (self.remote_logdir, self.hash)
        self.remote_namelist = "%s/refinement_namelist_%i" % (self.remote_namelist_dir, self.hash)
        self.local_namelist = "NewWorlds/namelist_precisions_%i" % self.hash

        # parameters to replace:
        self.parameters = {"JOBNAME": self.job_name,
                           "RUNDIR": self.remote_rundir,
                           "LOGDIR": self.remote_logdir,
                           "NAMELIST_PATH": self.remote_namelist,
                           "NUMPROC": "48"
                           }
        self.jobscript = generate_jobscript(self.template, self.parameters)

    def compare_with_job(self, other_job, outputs_file_name="", script=""):
        folder_files_2 = other_job.communicator.list_dir(other_job.remote_rundir)
        filename_2 = [f for f in folder_files_2 if re.search(outputs_file_name, f)][0]
        results_2 = "%s/%s" % (other_job.remote_rundir, filename_2)

        return self.compare_with_file(results_2, outputs_file_name, script)

    def compare_with_file(self, file_path, outputs_file_name="", script=""):
        folder_files_1 = self.communicator.list_dir(self.remote_rundir)
        filename_1 = [f for f in folder_files_1 if re.search(outputs_file_name, f)][0]
        results_1 = "%s/%s" % (self.remote_rundir, filename_1)

        results_2 = file_path

        with open(script) as input_script:
            remote_script = input_script.read()

        remote_script = remote_script.replace("%FILENAME1%", results_1)
        remote_script = remote_script.replace("%FILENAME2%", results_2)

        local_wrapper_path = "/home/Earth/otintopr/RPE-Canal/AutoRPE/EvaluationScripts/wrapper.sh"
        remote_wrapper_path = "%s/python_wrapper.sh" % self.remote_logdir

        self.communicator.put(local_wrapper_path, remote_wrapper_path)

        remote_script_name = "compare_script.py"
        remote_script_path = "%s/%s" % (self.remote_logdir, remote_script_name)
        with self.communicator.sftp.file(remote_script_path, "w") as remote_script_file:
            remote_script_file.write(remote_script)

        # Change permissions for the wrapper
        command = "chmod 777 %s" % remote_wrapper_path
        stdin, stdout, stderr = self.communicator.execute(command)
        output = str(stdout.read())

        # Execute script through wrapper
        command = "%s %s" % (remote_wrapper_path, remote_script_path)
        stdin, stdout, stderr = self.communicator.execute(command)
        output = str(stdout.read().decode())
        err = str(stderr.read().decode())

        if output.count("not found!"):
            print("File not found")
            return None

        output = [line for line in output.split("\n") if line.strip()]
        results = {}
        for line in output:
            name, value = line.split()
            value = float(value)
            results[name] = value

        if err.strip():
            print(err)
        return results


class BinarySearchJob(Job):
    def __init__(self, variable_set, reduced_precision_level=10, communicator=None,
                 vault=None, template="", assertion_template=""):
        super().__init__(variable_set, reduced_precision_level=reduced_precision_level,
                         communicator=communicator, vault=vault, template=template)
        # This class extends the classical job putting some extra variables that define the relations
        self.parent = None
        self.child = []
        self.child_number = None
        self.level = None
        self.banned_variables = []
        self.thresholds = {}
        if assertion_template:
            self.asserting = True
            self.assertion_template = assertion_template
        else:
            self.asserting = False
            self.assertion_template = ""

    def identifier(self):
        return var_set_identifier(self)

    def evaluate_success(self):
        success = True
        self.result()
        try:
            for key in self.thresholds.keys():
                if self._result[key] >= self.thresholds[key]:
                    success = False
        except TypeError:
            success = False
        return success

    def normalized_error(self):
        values = self.result()
        error = 0
        if values is None:
            return None
        for key in values:
            error += (values[key]/self.thresholds[key])**2
        return (error/len(values))**.5

    def ancestors(self):
        to_return = []
        cv = self
        while cv.parent is not None:
            to_return.append(cv.parent)
            cv = cv.parent
        # return list(reversed(to_return))
        return list(to_return)

    def descendants(self):
        to_return = []
        pre_list = [child for child in self.child]
        while len(pre_list) > 0:
            cv = pre_list.pop(0)
            for child in cv.child:
                if child not in pre_list:
                    pre_list.append(child)
            to_return.append(cv)
        return to_return

    def spawn_children(self):
        if len(self.variable_set) >= 2:
            if self.child == []:
                child_thresholds = dict(self.thresholds)
                for key in child_thresholds:
                    factor = (3. / 4.)
                    child_thresholds[key] = child_thresholds[key] * factor
                _set1, _set2 = divide_set(self.variable_set)
                for index, var_set in enumerate([_set1, _set2]):
                    job = BinarySearchJob(var_set, reduced_precision_level=self.reduced_precision_level,
                                          communicator=self.communicator, vault=self.vault, template=self.template,
                                          assertion_template=self.assertion_template)
                    job.parent = self
                    job.child_number = index
                    job.level = self.level + 1
                    job.thresholds = dict(child_thresholds)
                    self.child.append(job)
                return True

        return False

    def driver(self, queue, max_running_jobs):
        if self.status == "PENDING":
            if len(queue.running) < max_running_jobs:
                self.submit()
                self.status = "RUNNING"
                queue.pending.remove(self)
                queue.running.append(self)
        elif self.status == "RUNNING":
            self.check_running(queue)
        elif self.status == "ASSERTION_PENDING":
            self.check_assertion(queue)
        elif self.status == "SUSPENDED":
            self.check_child(queue)

    def check_running(self, queue):
        status = self.remote_status
        if status == "COMPLETED":
            success = self.evaluate_success()
            if success:
                self.qualify(queue)
            else:
                self.fail(queue)

        elif status == "FAILED":
            self.fail(queue)
        elif status == "RUNNING":
            pass
        elif status == "PENDING":
            pass
        elif status == "TIMEOUT":
            self.submit(force=True)
        else:
            raise AssertionError("Unhandled status %s" % status)

    def move_to_success(self, queue):
        queue[self.status].remove(self)
        self.status = "SUCCESS"
        queue.success.append(self)

    def qualify(self, queue, from_assertion=False):
        if from_assertion:
            self.move_to_success(queue)
        else:
            if self.asserting:
                self.double_check_result(queue)
            else:
                self.move_to_success(queue)

    def fail(self, queue):
        if self.spawn_children():
            queue[self.status].remove(self)
            self.status = "SUSPENDED"
            queue.suspended.append(self)

            # If possible, add the child to que queue
            for child in self.child:
                # If the set is already banned, it is not necessary to run it
                queue.pending.append(child)
        # If the set can not be split, ban the set
        else:
            queue[self.status].remove(self)
            self.status = "FAILED"
            queue.failed.append(self)

    def check_assertion(self, queue):
        job = self.assertion_job
        status = job.remote_status
        if status == "COMPLETED":
            success = job.evaluate_success()
            if success:
                self.qualify(queue, from_assertion=True)
            else:
                self.fail(queue)
        elif status == "FAILED":
            self.fail(queue)

    def double_check_result(self, queue):
        if self.normalized_error() > 0:
            # Launching Assertion Job!
            assertion_job = AssertionJob(self.variable_set[:],
                                         reduced_precision_level=self.reduced_precision_level,
                                         communicator=self.communicator, vault=self.vault,
                                         template=self.assertion_template, year=1, member=3,
                                         original_job=self)
            self.assertion_job = assertion_job
            assertion_job.submit()
            queue[self.status].remove(self)
            self.status = "ASSERTION_PENDING"
            queue.assertion_pending.append(self)
        else:
            self.qualify(queue, from_assertion=True)

    def check_child(self, queue):
        statuses = list(sorted([c.status for c in self.child]))
        if statuses == ["FAILED", "SUCCESS"]:
            pass

        all_finished = all([c.status == "FAILED" or c.status == "SUCCESS" for c in self.child])
        if all_finished:
            for child in self.child:
                self.banned_variables += child.banned_variables
                if child.status == "FAILED":
                    self.banned_variables += child.variable_set
            self.variable_set = [var for var in self.variable_set if var not in self.banned_variables]

            if self.variable_set is []:
                self.fail(queue)
            else:
                self.status = "PENDING"
                queue.pending.append(self)
                queue.suspended.remove(self)


class AssertionJob(Job):
    def __init__(self, variable_set=[], reduced_precision_level=10, communicator=None, vault=None,
                 template="Templates/singleJobscript.sh", year=1, member=0, hash_prefix="", original_job=None):
        self.year = year
        self.member = member
        self.hash_prefix = hash_prefix
        self.original_job = original_job
        super().__init__(variable_set, reduced_precision_level=reduced_precision_level,
                         communicator=communicator, vault=vault, template=template)

    def update_parameters(self):
        super().update_parameters()
        # parameters to replace:
        self.parameters = {"JOBNAME": self.job_name,
                           "RUNDIR": self.remote_rundir,
                           "LOGDIR": self.remote_logdir,
                           "NAMELIST_PATH": self.remote_namelist,
                           "NUMPROC": "48",
                           "YEAR": "%02i" % self.year,
                           "MEMBER": "%i" % self.member
                           }
        self.jobscript = generate_jobscript(self.template, self.parameters)

    @property
    def hash(self):
        variable_set_part = str(hash_from_list(self.variable_set))
        year_and_member_part = "%s%02i%i" % (self.hash_prefix, self.year, self.member)
        final_hash = int(year_and_member_part+variable_set_part)
        return final_hash

    def compare_with_reference(self, outputs_file_name="ORCA2_1d_[0-9]+_[0-9]+_grid_T", script="EvaluationScripts/single_file.py"):
        self.update_parameters()
        results_folder = "/gpfs/scratch/pr1egi00/pr1egi23/MP-Repo/EnsemblesN4/results/year_%02i_ic%i/" % (
                         self.year, self.member)
        files = self.communicator.list_dir(results_folder)
        results_filename = [f for f in files if re.match(outputs_file_name, f)][0]
        results_filepath = "%s/%s" % (results_folder, results_filename)
        return self.compare_with_file(results_filepath, outputs_file_name=outputs_file_name, script=script)

    def identifier(self):
        if self.original_job:
            return self.original_job.identifier()

    @property
    def level(self):
        return self.original_job.level

    def check_if_result_already_exists(self, last_time_step=150):
        self.update_parameters()
        time_step_filename = "%s/time.step" % self.remote_rundir
        try:
            with self.communicator.sftp.open(time_step_filename) as f:
                data = int(f.read())
                if data != last_time_step:
                    return False
            self.recover_job_id()
            return True
        except IOError:
            return False

    def evaluate_success(self):
        return evaluate_success(self.original_job.thresholds, self.compare_with_reference(), 1)


class Queue(object):
    def __init__(self):
        # The list of possible status shall cover all the possible situations in which a job can be found.
        # PENDING: The job has been created and ready to be launched.
        # RUNNING: The job is RUNNING.
        # ASSERTION_PENDING: A Job that has been completed with successful results but still needs assertion.
        # SUCCESS: A job that has been completed with successful results and asserted.
        # FAILED: A job that has been completed with unsuccessful results or a failed assertion and that can not be
        # split ( Already was or a single variable set).
        # SUSPENDED: A job that has been completed with unsuccessful results that has been split and depends on the
        # results of descendant jobs.

        self.pending = []
        self.running = []
        self.assertion_pending = []
        self.success = []
        self.failed = []
        self.suspended = []
        self.__dict = {"PENDING": self.pending, "RUNNING": self.running, "ASSERTION_PENDING": self.assertion_pending,
                       "SUCCESS": self.success, "FAILED": self.failed, "SUSPENDED": self.suspended}

    def sort(self, lower_first=True):
        if lower_first:
            self.pending = list(sorted(self.pending, key=lambda _x: _x.level))
        else:
            self.pending = list(sorted(self.pending, key=lambda _x: - _x.level))

    def all(self):
        return self.running + self.assertion_pending + self.pending + self.success + self.failed + self.suspended

    def __getitem__(self, item):
        return self.__dict[item]
